from json import JSONEncoder
from typing import (
    TYPE_CHECKING,
    Type,
    cast,
)

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.utils.module_loading import import_string


if TYPE_CHECKING:
    from revy.context import Context
    from revy.contrib.django.models import (
        AbstractAttributeDelta,
        AbstractDelta,
        AbstractObjectDelta,
        AbstractRevision,
    )


__all__ = (
    'get_revision_model',
    'get_delta_model',
    'get_object_delta_model',
    'get_attribute_delta_model',
    'get_json_encoder_class',
    'get_context_class',
)


def _get_swappable_model(
    setting_attname: str,
    model_class_path: str,
) -> Type[Model]:
    model: Type[Model]
    try:
        model = cast(
            Type[Model],
            apps.get_model(
                model_class_path,
                require_ready=False,
            ),
        )
    except ValueError:
        raise ImproperlyConfigured(
            f"{setting_attname} must be of the form 'app_label.model_name'",
        )
    except LookupError:
        raise ImproperlyConfigured(
            f"{setting_attname} refers to model '{model_class_path}' that has not been installed"
        )
    return model


def get_revision_model() -> Type['AbstractRevision']:
    from revy.contrib.django.conf import settings
    return cast(
        Type['AbstractRevision'],
        _get_swappable_model(
            settings.REVISION_MODEL_ATTNAME,
            settings.REVISION_MODEL,
        ),
    )


def get_delta_model() -> Type['AbstractDelta']:
    from revy.contrib.django.conf import settings
    return cast(
        Type['AbstractDelta'],
        _get_swappable_model(
            settings.DELTA_MODEL_ATTNAME,
            settings.DELTA_MODEL,
        ),
    )


def get_object_delta_model() -> Type['AbstractObjectDelta']:
    from revy.contrib.django.conf import settings
    return cast(
        Type['AbstractObjectDelta'],
        _get_swappable_model(
            settings.OBJECT_DELTA_MODEL_ATTNAME,
            settings.OBJECT_DELTA_MODEL,
        ),
    )


def get_attribute_delta_model() -> Type['AbstractAttributeDelta']:
    from revy.contrib.django.conf import settings
    return cast(
        Type['AbstractAttributeDelta'],
        _get_swappable_model(
            settings.ATTRIBUTE_DELTA_MODEL_ATTNAME,
            settings.ATTRIBUTE_DELTA_MODEL,
        ),
    )


def get_json_encoder_class() -> Type[JSONEncoder]:
    from revy.contrib.django.conf import settings
    return cast(
        Type[JSONEncoder],
        import_string(settings.JSON_ENCODER_CLASS),
    )


def get_context_class() -> Type['Context']:
    from revy.contrib.django.conf import settings
    return cast(
        Type['Context'],
        import_string(settings.CONTEXT_CLASS),
    )



from revy.contrib.django.setup import setup
from revy.contrib.django.utils import (
    get_attribute_delta_model,
    get_context_class,
    get_delta_model,
    get_object_delta_model,
    get_revision_model,
)


__all__ = (
    'setup',
    'get_revision_model',
    'get_delta_model',
    'get_object_delta_model',
    'get_attribute_delta_model',
    'get_context_class',
)



from revy.__version__ import (
    VERSION,
    __version__,
)
from revy.context import Context
from revy.globals import global_storage
from revy.storages import (
    ASGIRefLocalStorage,
    Storage,
)


__all__ = (
    '__version__',
    'VERSION',
    'Context',
    'global_storage',
    'Storage',
    'ASGIRefLocalStorage',
)



from revy.abc.attribute_delta import AttributeDelta
from revy.abc.delta import Delta
from revy.abc.lazy_revision import LazyRevision
from revy.abc.object_delta import ObjectDelta
from revy.abc.revision import Revision


__all__ = (
    'AttributeDelta',
    'Delta',
    'ObjectDelta',
    'Revision',
    'LazyRevision',
)



import dataclasses
from typing import (
    Any,
    Dict,
)


__all__ = (
    'METADATA_NAME',
    'wrap_metadata',
    'unwrap_metadata',
)


METADATA_NAME = 'revy'


def wrap_metadata(
    **kwargs: Any
) -> Dict[str, Any]:
    return {
        METADATA_NAME: kwargs,
    }


def unwrap_metadata(
    field: dataclasses.Field,
) -> Dict[str, Any]:
    metadata = field.metadata.get(METADATA_NAME, dict())
    return metadata



from typing import Sequence

from django.conf import settings


__all__ = (
    'MODELS_ATTNAME',
    'DEFAULT_MODELS',
    'MODELS',
    'REVISION_MODEL_ATTNAME',
    'DEFAULT_REVISION_MODEL',
    'REVISION_MODEL',
    'INCLUDE_AUTO_CREATED_MODELS_ATTNAME',
    'DEFAULT_INCLUDE_AUTO_CREATED_MODELS',
    'INCLUDE_AUTO_CREATED_MODELS',
    'REVISION_MODEL_ATTNAME',
    'DEFAULT_REVISION_MODEL',
    'REVISION_MODEL',
    'DELTA_MODEL_ATTNAME',
    'DEFAULT_DELTA_MODEL',
    'DELTA_MODEL',
    'OBJECT_DELTA_MODEL_ATTNAME',
    'DEFAULT_OBJECT_DELTA_MODEL',
    'OBJECT_DELTA_MODEL',
    'ATTRIBUTE_DELTA_MODEL_ATTNAME',
    'DEFAULT_ATTRIBUTE_DELTA_MODEL',
    'ATTRIBUTE_DELTA_MODEL',
    'JSON_ENCODER_CLASS_ATTNAME',
    'DEFAULT_JSON_ENCODER_CLASS',
    'JSON_ENCODER_CLASS',
    'DELETION_ITERATOR_CHUNK_SIZE_ATTNAME',
    'DEFAULT_DELETION_ITERATOR_CHUNK_SIZE',
    'DELETION_ITERATOR_CHUNK_SIZE',
    'CONTEXT_CLASS_ATTNAME',
    'DEFAULT_CONTEXT_CLASS',
    'CONTEXT_CLASS',
)


MODELS_ATTNAME = 'REVY_MODELS'

DEFAULT_MODELS = ('*',)

MODELS: Sequence[str]


INCLUDE_AUTO_CREATED_MODELS_ATTNAME = 'REVY_INCLUDE_AUTO_CREATED_MODELS'

DEFAULT_INCLUDE_AUTO_CREATED_MODELS = True

INCLUDE_AUTO_CREATED_MODELS: bool


REVISION_MODEL_ATTNAME = 'REVY_REVISION_MODEL'

DEFAULT_REVISION_MODEL = 'revy.Revision'

REVISION_MODEL: str


DELTA_MODEL_ATTNAME = 'REVY_DELTA_MODEL'

DEFAULT_DELTA_MODEL = 'revy.Delta'

DELTA_MODEL: str


OBJECT_DELTA_MODEL_ATTNAME = 'REVY_OBJECT_DELTA_MODEL'

DEFAULT_OBJECT_DELTA_MODEL = 'revy.ObjectDelta'

OBJECT_DELTA_MODEL: str


ATTRIBUTE_DELTA_MODEL_ATTNAME = 'REVY_ATTRIBUTE_DELTA_MODEL'

DEFAULT_ATTRIBUTE_DELTA_MODEL = 'revy.AttributeDelta'

ATTRIBUTE_DELTA_MODEL: str


JSON_ENCODER_CLASS_ATTNAME = 'REVY_JSON_ENCODER_CLASS'

DEFAULT_JSON_ENCODER_CLASS = 'django.core.serializers.json.DjangoJSONEncoder'

JSON_ENCODER_CLASS: str


DELETION_ITERATOR_CHUNK_SIZE_ATTNAME = 'REVY_DELETION_ITERATOR_CHUNK_SIZE'

DEFAULT_DELETION_ITERATOR_CHUNK_SIZE = 1000

DELETION_ITERATOR_CHUNK_SIZE: int


CONTEXT_CLASS_ATTNAME = 'REVY_CONTEXT_CLASS'

DEFAULT_CONTEXT_CLASS = 'revy.Context'

CONTEXT_CLASS: str


def reload() -> None:
    global MODELS
    MODELS = getattr(
        settings,
        MODELS_ATTNAME,
        (),
    ) or DEFAULT_MODELS
    if not hasattr(settings, MODELS_ATTNAME):
        setattr(settings, MODELS_ATTNAME, MODELS)

    global INCLUDE_AUTO_CREATED_MODELS
    INCLUDE_AUTO_CREATED_MODELS = getattr(
        settings,
        INCLUDE_AUTO_CREATED_MODELS_ATTNAME,
        DEFAULT_INCLUDE_AUTO_CREATED_MODELS,
    )
    if not hasattr(settings, INCLUDE_AUTO_CREATED_MODELS_ATTNAME):
        setattr(settings, INCLUDE_AUTO_CREATED_MODELS_ATTNAME, INCLUDE_AUTO_CREATED_MODELS)

    global REVISION_MODEL
    REVISION_MODEL = getattr(
        settings,
        REVISION_MODEL_ATTNAME,
        None,
    ) or DEFAULT_REVISION_MODEL
    if not hasattr(settings, REVISION_MODEL_ATTNAME):
        setattr(settings, REVISION_MODEL_ATTNAME, REVISION_MODEL)

    global DELTA_MODEL
    DELTA_MODEL = getattr(
        settings,
        DELTA_MODEL_ATTNAME,
        None,
    ) or DEFAULT_DELTA_MODEL
    if not hasattr(settings, DELTA_MODEL_ATTNAME):
        setattr(settings, DELTA_MODEL_ATTNAME, DELTA_MODEL)

    global OBJECT_DELTA_MODEL
    OBJECT_DELTA_MODEL = getattr(
        settings,
        OBJECT_DELTA_MODEL_ATTNAME,
        None,
    ) or DEFAULT_OBJECT_DELTA_MODEL
    if not hasattr(settings, OBJECT_DELTA_MODEL_ATTNAME):
        setattr(settings, OBJECT_DELTA_MODEL_ATTNAME, OBJECT_DELTA_MODEL)

    global ATTRIBUTE_DELTA_MODEL
    ATTRIBUTE_DELTA_MODEL = getattr(
        settings,
        ATTRIBUTE_DELTA_MODEL_ATTNAME,
        None,
    ) or DEFAULT_ATTRIBUTE_DELTA_MODEL
    if not hasattr(settings, ATTRIBUTE_DELTA_MODEL_ATTNAME):
        setattr(settings, ATTRIBUTE_DELTA_MODEL_ATTNAME, ATTRIBUTE_DELTA_MODEL)

    global JSON_ENCODER_CLASS
    JSON_ENCODER_CLASS = getattr(
        settings,
        JSON_ENCODER_CLASS_ATTNAME,
        None,
    ) or DEFAULT_JSON_ENCODER_CLASS
    if not hasattr(settings, JSON_ENCODER_CLASS_ATTNAME):
        setattr(settings, JSON_ENCODER_CLASS_ATTNAME, JSON_ENCODER_CLASS)

    global DELETION_ITERATOR_CHUNK_SIZE
    DELETION_ITERATOR_CHUNK_SIZE = getattr(
        settings,
        DELETION_ITERATOR_CHUNK_SIZE_ATTNAME,
        None,
    ) or DEFAULT_DELETION_ITERATOR_CHUNK_SIZE
    if not hasattr(settings, DELETION_ITERATOR_CHUNK_SIZE_ATTNAME):
        setattr(settings, DELETION_ITERATOR_CHUNK_SIZE_ATTNAME, DELETION_ITERATOR_CHUNK_SIZE)

    global CONTEXT_CLASS
    CONTEXT_CLASS = getattr(
        settings,
        CONTEXT_CLASS_ATTNAME,
        None,
    ) or DEFAULT_CONTEXT_CLASS
    if not hasattr(settings, CONTEXT_CLASS_ATTNAME):
        setattr(settings, CONTEXT_CLASS_ATTNAME, CONTEXT_CLASS)


reload()



from typing import (
    Any,
    Optional,
)

from revy.abc.delta import Delta
from revy.abc.object_delta import ObjectDelta


__all__ = (
    'AttributeDelta',
)


class AttributeDelta(Delta):  # noqa

    def get_object_delta(self) -> ObjectDelta:
        raise NotImplementedError()

    def set_object_delta(
        self,
        object_delta: ObjectDelta,
    ) -> None:
        raise NotImplementedError()

    def get_attribute_name(self) -> str:
        raise NotImplementedError()

    def set_attribute_name(
        self,
        attribute_name: str,
    ) -> None:
        raise NotImplementedError()

    def get_old_value(self) -> Optional[Any]:
        raise NotImplementedError()

    def set_old_value(
        self,
        old_value: Optional[Any],
    ) -> None:
        raise NotImplementedError()

    def get_new_value(self) -> Optional[Any]:
        raise NotImplementedError()

    def set_new_value(
        self,
        new_value: Optional[Any],
    ) -> None:
        raise NotImplementedError()



from typing import (
    Any,
    Iterable,
    TYPE_CHECKING,
)

from revy.abc.delta import Delta


if TYPE_CHECKING:
    from revy.abc.attribute_delta import AttributeDelta


__all__ = (
    'ObjectDelta',
)


class ObjectDelta(Delta):  # noqa

    def get_actors(self) -> Iterable[Any]:
        raise NotImplementedError()

    def get_attribute_deltas(self) -> Iterable['AttributeDelta']:
        raise NotImplementedError()



import ast
from pathlib import Path
from typing import (
    Optional,
    cast,
)


BASE_DIR = Path(__file__).resolve().parent.parent.parent.parent

SWAPPABLE_SAMPLE_DIR = BASE_DIR / 'samples' / 'django' / 'swappable'

INITIAL_MIGRATION_FILE_PATH = SWAPPABLE_SAMPLE_DIR / 'revisions' / 'migrations' / '0001_initial.py'


def main():
    migration_file_content = INITIAL_MIGRATION_FILE_PATH.read_text()
    migration_module = ast.parse(migration_file_content)
    migration_class_def: Optional[ast.ClassDef] = None
    for node in migration_module.body:
        if isinstance(node, ast.ClassDef) and node.name == 'Migration':
            migration_class_def = node
            break
    if migration_class_def is None:
        return
    for node in migration_class_def.body:
        if isinstance(node, ast.Assign):
            name = cast(ast.Name, node.targets[0])
            if name.id == 'replaces':
                return
    replaces_node = ast.Assign(
        lineno=-1,
        targets=[
            ast.Name(
                id='replaces',
                ctx=ast.Store(),
            ),
        ],
        value=ast.List(
            elts=[
                ast.Tuple(
                    elts=[
                        ast.Constant(value='revy'),
                        ast.Constant(value='0001_initial'),
                    ],
                ),
            ],
        ),
    )
    migration_class_def.body.insert(0, replaces_node)
    INITIAL_MIGRATION_FILE_PATH.write_text(
        ast.unparse(migration_module),
    )


if __name__ == '__main__':
    main()



from django.apps import AppConfig
from django.core.signals import setting_changed


__all__ = (
    'RevyConfig',
)


class RevyConfig(AppConfig):

    name = 'revy.contrib.django'

    verbose_name = 'Revy'

    label = 'revy'

    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self) -> None:
        from revy.contrib.django import setup
        setup()

        setting_changed.connect(
            setup,
            dispatch_uid='revy.contrib.django.setup',
        )



from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

import stackholm

from revy.abc import (
    LazyRevision,
    Revision,
)


__all__ = (
    'ContextMeta',
    'Context',
)


class ContextMeta(type):

    def __new__(
        mcs,
        name: str,
        bases: Tuple[Type, ...],
        namespace: Dict[str, Any],
    ) -> type:
        if namespace.get('_storage') is None:
            from revy.globals import global_storage
            namespace['_storage'] = global_storage
        return super(ContextMeta, mcs).__new__(mcs, name, bases, namespace)


class Context(
    stackholm.Context,
    metaclass=ContextMeta,
):

    class Key:

        IS_DISABLED = 'is_disabled'

        REVISION = 'revision'

        REVISION_DESCRIPTION = 'revision_description'

        ACTOR = 'actor'

        OBJECT_DELTA_DESCRIPTION = 'object_delta_description'

        ATTRIBUTE_DELTA_DESCRIPTION = 'attribute_delta_description'

        DELETION_DESCRIPTION = 'deletion_description'

    @classmethod
    def via(
        cls,
        **kwargs: Any,
    ) -> 'Context':
        context = cls()
        context.checkpoint_data.update(kwargs)
        return context

    @classmethod
    def disable(cls) -> None:
        cls.set_checkpoint_value(cls.Key.IS_DISABLED, True)

    @classmethod
    def is_disabled(cls) -> bool:
        if cls.get_current() is None:
            return True
        if cls.get_checkpoint_value(cls.Key.IS_DISABLED, False):
            return True
        return False

    @classmethod
    def as_disabled(cls) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.IS_DISABLED] = True
        return context

    @classmethod
    def enable(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.IS_DISABLED)

    @classmethod
    def is_enabled(cls) -> bool:
        return not cls.is_disabled()

    @classmethod
    def as_enabled(cls) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.IS_DISABLED] = False
        return context

    @classmethod
    def get_revision(cls) -> Optional[Revision]:
        revision = cls.get_checkpoint_value(cls.Key.REVISION)
        if isinstance(revision, LazyRevision):
            return revision()
        return revision

    @classmethod
    def set_revision(
        cls,
        revision: Union[
            None,
            Revision,
            LazyRevision,
        ],
    ) -> None:
        cls.set_checkpoint_value(
            cls.Key.REVISION,
            revision,
        )

    @classmethod
    def unset_revision(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.REVISION)

    @classmethod
    def reset_revision(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.REVISION)

    @classmethod
    def via_revision(
        cls,
        revision: Union[
            None,
            Revision,
            LazyRevision,
        ],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.REVISION] = revision
        return context

    @classmethod
    def get_revision_description(cls) -> str:
        return cls.get_checkpoint_value(cls.Key.REVISION_DESCRIPTION) or ''

    @classmethod
    def set_revision_description(
        cls,
        revision_description: Optional[str],
    ) -> None:
        cls.set_checkpoint_value(
            cls.Key.REVISION_DESCRIPTION,
            revision_description,
        )

    @classmethod
    def unset_revision_description(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.REVISION_DESCRIPTION)

    @classmethod
    def reset_revision_description(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.REVISION_DESCRIPTION)

    @classmethod
    def via_revision_description(
        cls,
        revision_description: Optional[str],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.REVISION_DESCRIPTION] = revision_description
        return context

    @classmethod
    def get_actor(cls) -> Optional[Any]:
        return cls.get_checkpoint_value(cls.Key.ACTOR)

    @classmethod
    def set_actor(
        cls,
        actor: Optional[Any],
    ) -> None:
        cls.set_checkpoint_value(cls.Key.ACTOR, actor)

    @classmethod
    def unset_actor(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.ACTOR)

    @classmethod
    def reset_actor(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.ACTOR)

    @classmethod
    def via_actor(
        cls,
        actor: Optional[Any],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.ACTOR] = actor
        return context

    @classmethod
    def get_object_delta_description(cls) -> str:
        return cls.get_checkpoint_value(cls.Key.OBJECT_DELTA_DESCRIPTION) or ''

    @classmethod
    def set_object_delta_description(
        cls,
        object_delta_description: Optional[str],
    ) -> None:
        cls.set_checkpoint_value(
            cls.Key.OBJECT_DELTA_DESCRIPTION,
            object_delta_description,
        )

    @classmethod
    def unset_object_delta_description(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.OBJECT_DELTA_DESCRIPTION)

    @classmethod
    def reset_object_delta_description(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.OBJECT_DELTA_DESCRIPTION)

    @classmethod
    def via_object_delta_description(
        cls,
        object_delta_description: Optional[str],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.OBJECT_DELTA_DESCRIPTION] = object_delta_description
        return context

    @classmethod
    def get_attribute_delta_description(cls) -> str:
        return cls.get_checkpoint_value(cls.Key.ATTRIBUTE_DELTA_DESCRIPTION) or ''

    @classmethod
    def set_attribute_delta_description(
        cls,
        attribute_delta_description: Optional[str],
    ) -> None:
        cls.set_checkpoint_value(
            cls.Key.ATTRIBUTE_DELTA_DESCRIPTION,
            attribute_delta_description,
        )

    @classmethod
    def unset_attribute_delta_description(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.ATTRIBUTE_DELTA_DESCRIPTION)

    @classmethod
    def reset_attribute_delta_description(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.ATTRIBUTE_DELTA_DESCRIPTION)

    @classmethod
    def via_attribute_delta_description(
        cls,
        attribute_delta_description: Optional[str],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.ATTRIBUTE_DELTA_DESCRIPTION] = attribute_delta_description
        return context

    @classmethod
    def get_deletion_description(cls) -> str:
        return cls.get_checkpoint_value(cls.Key.DELETION_DESCRIPTION) or ''

    @classmethod
    def set_deletion_description(
        cls,
        deletion_description: Optional[str],
    ) -> None:
        cls.set_checkpoint_value(
            cls.Key.DELETION_DESCRIPTION,
            deletion_description,
        )

    @classmethod
    def unset_deletion_description(cls) -> None:
        cls.pop_checkpoint_value(cls.Key.DELETION_DESCRIPTION)

    @classmethod
    def reset_deletion_description(cls) -> None:
        cls.reset_checkpoint_value(cls.Key.DELETION_DESCRIPTION)

    @classmethod
    def via_deletion_description(
        cls,
        deletion_description: Optional[str],
    ) -> 'Context':
        context = cls()
        context.checkpoint_data[cls.Key.DELETION_DESCRIPTION] = deletion_description
        return context



from typing import Tuple


__all__ = (
    'VersionTuple',
    'Version',
    'version',
    'VERSION',
    '__version__',
)


VersionTuple = Tuple[int, int, int, str, int]


class Version:

    major: int

    minor: int

    micro: int

    segment: str

    build: int

    def __init__(
        self,
        major: int,
        minor: int,
        micro: int,
        segment: str = 'final',
        build: int = 0,
    ) -> None:
        assert segment in ('alpha', 'beta', 'rc', 'final')  # noqa
        assert build >= 0  # noqa
        self.major = major
        self.minor = minor
        self.micro = micro
        self.segment = segment
        self.build = build

    def to_tuple(self) -> VersionTuple:
        return self.major, self.minor, self.micro, self.segment, self.build

    def __str__(self) -> str:
        segment = self.segment if self.segment != 'final' else ''
        build = str(self.build) if self.build > 0 else ''
        return f'{self.major}.{self.minor}.{self.micro}{segment}{build}'


version = Version(1, 0, 0, 'alpha', 16)

VERSION = version.to_tuple()

__version__ = str(version)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import datetime

# -- Project information -----------------------------------------------------
from typing import List


project = 'Revy'
copyright = f'{datetime.date.today().year}, Ertuğrul Noyan Keremoğlu'
author = 'Ertuğrul Noyan Keremoğlu'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

autodoc_inherit_docstrings = False

autodoc_preserve_defaults = True

autodoc_default_options = {
    'show-inheritance': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
}

autodoc_typehints = 'both'

autodoc_typehints_format = 'short'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

html_theme_options = {
    'use_edit_page_button': False,
    'search_bar_text': 'Search',
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/ertgl/revy',
            'icon': 'fab fa-github-square',
            'type': 'fontawesome'
        },
    ],
}

html_context = {
    "github_user": "ertgl",
    "github_repo": "revy",
    "github_version": "main",
    "doc_path": "docs/source",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



from django.conf import settings
import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import revy.abc.attribute_delta
import revy.abc.delta
import revy.abc.object_delta
import revy.abc.revision


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Delta',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('actor_id', models.TextField(blank=True, db_index=True, default=None, null=True, verbose_name='actor ID')),
                ('action', models.CharField(choices=[('Create', 'Create'), ('Update', 'Update'), ('Delete', 'Delete'), ('Set', 'Set'), ('Unset', 'Unset')], db_index=True, max_length=6, verbose_name='action')),
                ('description', models.TextField(blank=True, default='', verbose_name='description')),
                ('content_id', models.TextField(blank=True, db_index=True, default=None, null=True, verbose_name='content ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='updated at')),
                ('actor_type', models.ForeignKey(blank=True, db_column='actor_type_id', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='deltas_created_by', related_query_name='delta_created_by', to='contenttypes.contenttype', verbose_name='actor type')),
                ('content_type', models.ForeignKey(blank=True, db_column='content_type_id', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='deltas', related_query_name='delta', to='contenttypes.contenttype', verbose_name='content type')),
            ],
            options={
                'verbose_name': 'delta',
                'verbose_name_plural': 'deltas',
                'db_table': 'revy__deltas',
                'swappable': 'REVY_DELTA_MODEL',
            },
            bases=(revy.abc.delta.Delta, models.Model),
        ),
        migrations.CreateModel(
            name='Revision',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(blank=True, default='', verbose_name='description')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='updated at')),
            ],
            options={
                'verbose_name': 'revision',
                'verbose_name_plural': 'revisions',
                'db_table': 'revy__revisions',
                'swappable': 'REVY_REVISION_MODEL',
            },
            bases=(revy.abc.revision.Revision, models.Model),
        ),
        migrations.CreateModel(
            name='ObjectDelta',
            fields=[
                ('object_delta_ptr', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='object_delta', related_query_name='object_delta', serialize=False, to=settings.REVY_DELTA_MODEL, verbose_name='object delta pointer')),
            ],
            options={
                'verbose_name': 'object delta',
                'verbose_name_plural': 'object deltas',
                'db_table': 'revy__object_deltas',
                'swappable': 'REVY_OBJECT_DELTA_MODEL',
            },
            bases=(revy.abc.object_delta.ObjectDelta, 'revy.delta', revy.abc.delta.Delta, models.Model),
        ),
        migrations.AddField(
            model_name='delta',
            name='revision',
            field=models.ForeignKey(db_column='revision_id', on_delete=django.db.models.deletion.CASCADE, related_name='deltas', related_query_name='delta', to=settings.REVY_REVISION_MODEL, verbose_name='revision'),
        ),
        migrations.CreateModel(
            name='AttributeDelta',
            fields=[
                ('field_name', models.TextField(db_index=True, verbose_name='field name')),
                ('old_value', models.JSONField(blank=True, default=None, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True, verbose_name='old value')),
                ('new_value', models.JSONField(blank=True, default=None, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True, verbose_name='new value')),
                ('attribute_delta_ptr', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='attribute_delta', related_query_name='attribute_delta', serialize=False, to=settings.REVY_DELTA_MODEL, verbose_name='attribute delta')),
                ('parent', models.ForeignKey(db_column='parent_id', on_delete=django.db.models.deletion.CASCADE, related_name='children', related_query_name='child', to=settings.REVY_OBJECT_DELTA_MODEL, verbose_name='parent')),
            ],
            options={
                'verbose_name': 'attribute delta',
                'verbose_name_plural': 'attribute deltas',
                'db_table': 'revy__attribute_deltas',
                'swappable': 'REVY_ATTRIBUTE_DELTA_MODEL',
            },
            bases=(revy.abc.attribute_delta.AttributeDelta, 'revy.delta', revy.abc.delta.Delta, models.Model),
        ),
    ]



import unittest

import revy


class GlobalStorageTestCase(unittest.TestCase):

    def test_global_storage(self) -> None:
        self.assertIs(revy.Context._storage, revy.global_storage)
        self.assertIsInstance(revy.Context._storage, revy.Storage)
        self.assertIsInstance(revy.Context._storage, revy.ASGIRefLocalStorage)



from typing import (
    Optional,
    Protocol,
    runtime_checkable,
)

from revy.abc.revision import Revision


__all__ = (
    'LazyRevision',
)


@runtime_checkable
class LazyRevision(Protocol):

    def __call__(self) -> Optional[Revision]:
        ...



from contextlib import (
    ExitStack,
    suppress,
)
import dataclasses
import functools
import inspect
from typing import (
    Any,
    Callable,
    Optional,
    TYPE_CHECKING,
    Type,
    cast,
)

from django.core.exceptions import FieldDoesNotExist
from django.db import transaction
from django.db.models import (
    Field,
    Model,
)
from django.db.models.options import Options

from revy.contrib.django.utils import (
    get_attribute_delta_model,
    get_context_class,
    get_delta_model,
    get_object_delta_model,
    get_revision_model,
)


if TYPE_CHECKING:
    from revy.contrib.django.models import AbstractRevision  # noqa


__all__ = (
    'Patcher',
    'MethodPatchData',
)


@dataclasses.dataclass()
class MethodPatchData:

    type_: type = dataclasses.field(
        kw_only=True,
    )

    method_name: str = dataclasses.field(
        kw_only=True,
    )

    original_method: Callable[..., Any] = dataclasses.field(
        kw_only=True,
    )

    patched_method: Callable[..., Any] = dataclasses.field(
        kw_only=True,
    )


class Patcher:

    PATCH_DATA_ATTNAME = '__revy__patch_data'

    @classmethod
    def get_method_patch_data(
        cls,
        type_: type,
        method_name: str,
    ) -> Optional[MethodPatchData]:
        attribute = getattr(type_, method_name, None)
        patch_data = getattr(attribute, cls.PATCH_DATA_ATTNAME, None)
        if isinstance(patch_data, MethodPatchData):
            return patch_data
        return None

    @classmethod
    def is_method_patched(
        cls,
        type_: type,
        method_name: str,
    ) -> bool:
        return cls.get_method_patch_data(type_, method_name) is not None

    @classmethod
    def apply_method_patch(
        cls,
        method_patch_data: MethodPatchData,
    ) -> None:
        setattr(
            method_patch_data.type_,
            method_patch_data.method_name,
            method_patch_data.patched_method,
        )
        setattr(
            method_patch_data.patched_method,
            cls.PATCH_DATA_ATTNAME,
            method_patch_data,
        )

    @classmethod
    def unpatch_method(
        cls,
        type_: type,
        method_name: str,
    ) -> None:
        method_patch_data = cls.get_method_patch_data(type_, method_name)
        if method_patch_data is None:
            return
        setattr(
            method_patch_data.type_,
            method_patch_data.method_name,
            method_patch_data.original_method,
        )
        delattr(
            method_patch_data.patched_method,
            cls.PATCH_DATA_ATTNAME,
        )

    @classmethod
    def patch_model(
        cls,
        model: Type[Model],
    ) -> None:
        excluded_models = (
            get_revision_model(),
            get_delta_model(),
            get_object_delta_model(),
            get_attribute_delta_model(),
        )
        if issubclass(model, excluded_models):
            return
        cls.patch_model_init(model)
        cls.patch_model_setattr(model)
        cls.patch_model_save_base(model)
        cls.patch_model_refresh_from_db(model)
        cls.patch_model_delete(model)

    @classmethod
    def unpatch_model(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_model_init(model)
        cls.unpatch_model_setattr(model)
        cls.unpatch_model_save_base(model)
        cls.unpatch_model_refresh_from_db(model)
        cls.unpatch_model_delete(model)

    @classmethod
    def patch_model_init(
        cls,
        model: Type[Model],
    ) -> None:

        if cls.is_method_patched(model, '__init__'):
            return

        from revy.contrib.django.models import get_model_instance_state

        context_class = get_context_class()

        original_init = cast(
            Callable[..., None],
            model.__init__,
        )

        @functools.wraps(original_init)
        def patched_init(
            self: Model,
            *args: Any,
            **kwargs: Any,
        ) -> None:

            if context_class.is_disabled():
                return original_init(self, *args, **kwargs)

            current_frame = inspect.currentframe()
            previous_frame = current_frame.f_back if current_frame else None
            caller_name = previous_frame.f_code.co_name if previous_frame else None
            caller_locals = previous_frame.f_locals if previous_frame else dict()
            caller_class = caller_locals.get('cls', self.__class__)
            caller_class_is_model = inspect.isclass(caller_class) and issubclass(caller_class, Model)
            caller_class_method_is_from_db = caller_class_is_model and caller_name == 'from_db'

            state = get_model_instance_state(self)

            state.init_keys = list(kwargs.keys())

            state.is_initialized = False

            state.is_new = not caller_class_method_is_from_db

            state.is_being_saved = False
            state.is_saved = False

            state.is_being_fetched = caller_class_method_is_from_db
            state.is_fetched = False

            state.is_being_deleted = False
            state.is_deleted = False

            original_init(self, *args, **kwargs)

            state.is_initialized = True

            state.is_being_fetched = False
            state.is_fetched = not state.is_new

        method_patch_data = MethodPatchData(
            type_=model,
            method_name='__init__',
            original_method=original_init,
            patched_method=patched_init,
        )

        cls.apply_method_patch(method_patch_data)

    @classmethod
    def unpatch_model_init(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_method(model, '__init__')

    @classmethod
    def patch_model_setattr(
        cls,
        model: Type[Model],
    ) -> None:

        if cls.is_method_patched(model, '__setattr__'):
            return

        from revy.contrib.django.models import get_model_instance_state

        context_class = get_context_class()

        original_setattr = cast(
            Callable[..., None],
            model.__setattr__,
        )

        @functools.wraps(original_setattr)
        def patched_setattr(
            self: Model,
            attname: str,
            value: Any,
        ) -> None:

            if context_class.is_disabled():
                return original_setattr(self, attname, value)

            meta = cast(Options, model._meta)  # noqa

            field: Optional[Field] = None
            with suppress(FieldDoesNotExist):
                field = cast(Field, meta.get_field(attname))
            if field is None:
                return original_setattr(self, attname, value)
            if field.one_to_many or field.many_to_many:
                return original_setattr(self, attname, value)

            new_value = value
            if field.is_relation and isinstance(value, Model):
                related_model = cast(Type[Model], value.__class__)
                related_meta = cast(Options, related_model._meta)  # noqa
                related_pk = related_meta.pk
                if related_pk is None:
                    return original_setattr(self, attname, value)
                new_value = value.__dict__.get(related_pk.attname)

            state = get_model_instance_state(self)

            is_initial = field.attname in state.init_keys or field.name in state.init_keys
            is_caused_by_system = not state.is_initialized and not is_initial
            is_caused_by_system |= state.is_being_saved
            is_caused_by_system |= state.is_being_fetched
            is_caused_by_system |= state.is_being_deleted

            old_value = state.previous_values.get(field.attname)
            is_changed = old_value != new_value
            if is_changed:
                state.previous_values[field.attname] = new_value

            original_setattr(self, attname, value)

            if state.is_being_fetched:
                return

            if not is_changed:
                return

            attribute_delta_class = get_attribute_delta_model()

            with context_class():

                if is_caused_by_system:
                    context_class.set_actor(None)

                previous_attribute_delta_indexes = state.field_name_to_attribute_delta_index_mapping.get(
                    field.attname,
                    [],
                )

                previous_attribute_delta_index: Optional[int] = None
                with suppress(IndexError):
                    previous_attribute_delta_index = previous_attribute_delta_indexes[-1]

                previous_attribute_delta = (
                    state.attribute_deltas[previous_attribute_delta_index]
                    if previous_attribute_delta_index is not None
                    else None
                )

                previous_actor = (
                    previous_attribute_delta.get_actor()
                    if previous_attribute_delta is not None
                    else None
                )

                actor = context_class.get_actor()

                is_first_delta = state.is_new and len(previous_attribute_delta_indexes) == 0

                action = (
                    attribute_delta_class.ACTION_SET
                    if is_first_delta or new_value not in (None, False, '')
                    else attribute_delta_class.ACTION_UNSET
                )

                description = context_class.get_attribute_delta_description()

                was_default = state.is_new
                was_default &= previous_actor is None
                was_default &= (
                    field.default in [previous_attribute_delta.get_new_value(), None]
                    if previous_attribute_delta is not None
                    else False
                )

                was_caused_by_system = state.is_initialized and previous_actor is None
                was_caused_by_system &= not is_initial

                should_discard_new_attribute_delta = previous_attribute_delta is not None
                should_discard_new_attribute_delta &= (previous_actor is actor) or was_caused_by_system
                should_discard_new_attribute_delta |= was_default

                concrete_attribute_delta = (
                    previous_attribute_delta
                    if should_discard_new_attribute_delta
                    else attribute_delta_class()
                )

                assert concrete_attribute_delta is not None  # noqa

                concrete_attribute_delta.set_action(action)
                concrete_attribute_delta.set_description(description)
                concrete_attribute_delta.set_new_value(new_value)

                if not should_discard_new_attribute_delta:

                    concrete_attribute_delta.set_actor(actor)
                    concrete_attribute_delta.set_attribute_name(field.attname)
                    concrete_attribute_delta.set_old_value(old_value)

                    state.attribute_deltas.append(concrete_attribute_delta)
                    attribute_delta_index = len(state.attribute_deltas) - 1
                    state.field_name_to_attribute_delta_index_mapping[field.attname].append(
                        attribute_delta_index,
                    )

        method_patch_data = MethodPatchData(
            type_=model,
            method_name='__setattr__',
            original_method=original_setattr,
            patched_method=patched_setattr,
        )

        cls.apply_method_patch(method_patch_data)

    @classmethod
    def unpatch_model_setattr(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_method(model, '__setattr__')

    @classmethod
    def patch_model_save_base(
        cls,
        model: Type[Model],
    ) -> None:

        if cls.is_method_patched(model, 'save_base'):
            return

        from revy.contrib.django.models import get_model_instance_state

        context_class = get_context_class()

        original_save_base = cast(
            Callable[..., None],
            model.save_base,
        )

        @functools.wraps(original_save_base)
        def patched_save_base(
            self: Model,
            *args: Any,
            **kwargs: Any,
        ) -> None:

            if context_class.is_disabled():
                return original_save_base(self, *args, **kwargs)

            state = get_model_instance_state(self)

            snapshot = state.backup()

            def restore_state() -> None:
                state.restore(snapshot)

            was_new = state.is_new
            state.is_being_saved = True

            with ExitStack() as exit_stack, transaction.atomic():

                exit_stack.callback(restore_state)
                original_save_base(self, *args, **kwargs)
                exit_stack.pop_all()

                revision = context_class.get_revision()
                if revision is None:
                    revision_class = get_revision_model()
                    revision = revision_class()
                    revision.set_description(context_class.get_revision_description())
                    revision.save()
                revision = cast('AbstractRevision', revision)

                object_delta_class = get_object_delta_model()

                action = (
                    object_delta_class.ACTION_CREATE
                    if was_new
                    else object_delta_class.ACTION_UPDATE
                )

                object_delta = object_delta_class()
                object_delta.set_revision(revision)
                object_delta.set_actor(context_class.get_actor())
                object_delta.set_action(action)
                object_delta.set_description(context_class.get_object_delta_description())
                object_delta.set_object(self)
                object_delta.save()

                for attribute_delta in state.attribute_deltas:
                    attribute_delta.set_revision(revision)
                    attribute_delta.set_object_delta(object_delta)
                    attribute_delta.set_object(self)
                    attribute_delta.save()
                state.attribute_deltas.clear()
                state.field_name_to_attribute_delta_index_mapping.clear()

                state.is_new = False

                state.is_being_saved = False
                state.is_saved = True

                state.is_being_fetched = False
                state.is_fetched = False

                state.is_being_deleted = False
                state.is_deleted = False

        method_patch_data = MethodPatchData(
            type_=model,
            method_name='save_base',
            original_method=original_save_base,
            patched_method=patched_save_base,
        )

        cls.apply_method_patch(method_patch_data)

    @classmethod
    def unpatch_model_save_base(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_method(model, 'save_base')

    @classmethod
    def patch_model_refresh_from_db(
        cls,
        model: Type[Model],
    ) -> None:

        if cls.is_method_patched(model, 'refresh_from_db'):
            return

        from revy.contrib.django.models import get_model_instance_state

        context_class = get_context_class()

        original_refresh_from_db = cast(
            Callable[..., None],
            model.refresh_from_db,
        )

        @functools.wraps(original_refresh_from_db)
        def patched_refresh_from_db(
            self: Model,
            *args: Any,
            **kwargs: Any,
        ) -> None:

            if context_class.is_disabled():
                return original_refresh_from_db(self, *args, **kwargs)

            state = get_model_instance_state(self)

            snapshot = state.backup()

            def restore_state() -> None:
                state.restore(snapshot)

            state.is_being_fetched = True

            with ExitStack() as exit_stack:
                exit_stack.callback(restore_state)
                original_refresh_from_db(self, *args, **kwargs)
                exit_stack.pop_all()

                for attribute_delta in state.attribute_deltas:
                    attribute_delta.set_object(self)

                state.is_new = False

                state.is_being_saved = False
                state.is_saved = False

                state.is_being_fetched = False
                state.is_fetched = True

                state.is_being_deleted = False
                state.is_deleted = False

        method_patch_data = MethodPatchData(
            type_=model,
            method_name='refresh_from_db',
            original_method=original_refresh_from_db,
            patched_method=patched_refresh_from_db,
        )

        cls.apply_method_patch(method_patch_data)

    @classmethod
    def unpatch_model_refresh_from_db(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_method(model, 'refresh_from_db')

    @classmethod
    def patch_model_delete(
        cls,
        model: Type[Model],
    ) -> None:

        if cls.is_method_patched(model, 'delete'):
            return

        from revy.contrib.django.models import get_model_instance_state

        context_class = get_context_class()

        original_delete = cast(
            Callable[..., None],
            model.delete,
        )

        @functools.wraps(original_delete)
        def patched_delete(
            self: Model,
            *args: Any,
            **kwargs: Any,
        ) -> None:

            if context_class.is_disabled():
                return original_delete(self, *args, **kwargs)

            state = get_model_instance_state(self)

            snapshot = state.backup()

            def restore_state() -> None:
                state.restore(snapshot)

            state.is_being_deleted = True

            with ExitStack() as exit_stack, transaction.atomic():

                exit_stack.callback(restore_state)

                revision = context_class.get_revision()
                if revision is None:
                    revision_class = get_revision_model()
                    revision = revision_class()
                    revision.set_description(context_class.get_revision_description())
                    revision.save()
                revision = cast('AbstractRevision', revision)

                object_delta_class = get_object_delta_model()
                object_delta = object_delta_class()
                object_delta.set_revision(revision)
                object_delta.set_actor(context_class.get_actor())
                object_delta.set_action(object_delta_class.ACTION_DELETE)
                object_delta.set_description(context_class.get_object_delta_description())
                object_delta.set_object(self)
                object_delta.save()

                for attribute_delta in state.attribute_deltas:
                    attribute_delta.set_revision(revision)
                    attribute_delta.set_object(self)
                    attribute_delta.save()
                state.attribute_deltas.clear()
                state.field_name_to_attribute_delta_index_mapping.clear()

                original_delete(self, *args, **kwargs)
                exit_stack.pop_all()

                state.is_new = False

                state.is_being_saved = False
                state.is_saved = False

                state.is_being_fetched = False
                state.is_fetched = False

                state.is_being_deleted = False
                state.is_deleted = True

        method_patch_data = MethodPatchData(
            type_=model,
            method_name='delete',
            original_method=original_delete,
            patched_method=patched_delete,
        )

        cls.apply_method_patch(method_patch_data)

    @classmethod
    def unpatch_model_delete(
        cls,
        model: Type[Model],
    ) -> None:
        cls.unpatch_method(model, 'delete')



from collections import defaultdict
import dataclasses
import datetime
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    runtime_checkable,
)

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Model
from django.db.models.constants import LOOKUP_SEP
from django.utils.translation import gettext_lazy as _

import revy.abc
from revy.contrib.django import _typing
from revy.contrib.django.aggregates import ObjectSnapshot
from revy.contrib.django.conf import settings
from revy.contrib.django.deletion import (
    CASCADE,
    SET,
    SET_DEFAULT,
    SET_NULL,
)
from revy.contrib.django.utils import (
    get_attribute_delta_model,
    get_delta_model,
    get_json_encoder_class,
    get_object_delta_model,
)
from revy.dataclasses.metadata import (
    unwrap_metadata,
    wrap_metadata,
)


__all__ = (
    'ObjectSnapshot',
    'CASCADE',
    'SET',
    'SET_DEFAULT',
    'SET_NULL',
    'AbstractRevision',
    'AbstractLazyRevision',
    'BaseRevision',
    'Revision',
    'AbstractDelta',
    'BaseDelta',
    'Delta',
    'AbstractObjectDelta',
    'BaseObjectDelta',
    'ObjectDelta',
    'AbstractAttributeDelta',
    'BaseAttributeDelta',
    'AttributeDelta',
    'ModelInstanceState',
    'get_model_instance_state',
)


JSONEncoder = get_json_encoder_class()


class AbstractRevision(
    revy.abc.Revision,
    models.Model,
):

    DESCRIPTION_FIELD_NAME = 'description'

    CREATED_AT_FIELD_NAME = 'created_at'

    UPDATED_AT_FIELD_NAME = 'updated_at'

    class Meta:  # type: ignore[override]
        abstract = True

    def get_description(self) -> str:
        return getattr(self, self.__class__.DESCRIPTION_FIELD_NAME)

    def set_description(
        self,
        description: str,
    ) -> None:
        setattr(self, self.__class__.DESCRIPTION_FIELD_NAME, description)

    def get_deltas(self) -> _typing.QuerySet['AbstractDelta']:  # type: ignore[misc]
        delta_class = get_delta_model()
        deltas = delta_class.objects.filter(**{
            delta_class.REVISION_FIELD_NAME: self.pk,
        })
        return deltas

    def get_object_deltas(self) -> _typing.QuerySet['AbstractObjectDelta']:  # type: ignore[misc]
        delta_class = get_delta_model()
        object_delta_class = get_object_delta_model()
        q_lhs = LOOKUP_SEP.join([
            object_delta_class.PARENT_LINK_FIELD_NAME,
            delta_class.REVISION_FIELD_NAME,
        ])
        object_deltas = object_delta_class.objects.filter(**{
            q_lhs: self.pk,
        })
        return object_deltas

    def get_attribute_deltas(self) -> _typing.QuerySet['AbstractAttributeDelta']:  # type: ignore[misc]
        delta_class = get_delta_model()
        attribute_delta_class = get_attribute_delta_model()
        q_lhs = LOOKUP_SEP.join([
            attribute_delta_class.PARENT_LINK_FIELD_NAME,
            delta_class.REVISION_FIELD_NAME,
        ])
        attribute_deltas = attribute_delta_class.objects.filter(**{
            q_lhs: self.pk,
        })
        return attribute_deltas

    def get_actors(self) -> Iterable[Model]:
        seen: Set[Model] = set()
        for delta in self.get_deltas():
            actor = delta.get_actor()
            if actor is None:
                continue
            if actor in seen:
                continue
            yield actor
            seen.add(actor)

    def get_created_at(self) -> datetime.datetime:
        return getattr(self, self.__class__.CREATED_AT_FIELD_NAME)

    def set_created_at(
        self,
        created_at: datetime.datetime,
    ) -> None:
        setattr(self, self.__class__.CREATED_AT_FIELD_NAME, created_at)

    def get_updated_at(self) -> datetime.datetime:
        return getattr(self, self.__class__.UPDATED_AT_FIELD_NAME)

    def set_updated_at(
        self,
        updated_at: datetime.datetime,
    ) -> None:
        setattr(self, self.__class__.UPDATED_AT_FIELD_NAME, updated_at)


@runtime_checkable
class AbstractLazyRevision(
    revy.abc.LazyRevision,
    Protocol,
):

    def __call__(self) -> Optional[AbstractRevision]:
        ...


class BaseRevision(
    AbstractRevision,
    models.Model,
):

    description: _typing.TextField[
        str,
        str,
    ] = models.TextField(
        verbose_name=_('description'),
        blank=True,
        null=False,
        default='',
    )

    created_at: _typing.DateTimeField[
        datetime.datetime,
        datetime.datetime,
    ] = models.DateTimeField(
        verbose_name=_('created at'),
        auto_now_add=True,
        blank=True,
        null=False,
        db_index=True,
    )

    updated_at: _typing.DateTimeField[
        datetime.datetime,
        datetime.datetime,
    ] = models.DateTimeField(
        verbose_name=_('updated at'),
        auto_now=True,
        blank=True,
        null=False,
        db_index=True,
    )

    class Meta:  # type: ignore[override]
        abstract = True


class Revision(
    BaseRevision,
    models.Model,
):

    class Meta:  # type: ignore[override]

        verbose_name = _('revision')

        verbose_name_plural = _('revisions')

        db_table = 'revy__revisions'

        swappable = settings.REVISION_MODEL_ATTNAME


class AbstractDelta(
    revy.abc.Delta,
    models.Model,
):

    PARENT_LINK_FIELD_NAME = 'delta_ptr'

    REVISION_FIELD_NAME = 'revision'

    ACTOR_FIELD_NAME = 'actor'

    ACTION_FIELD_NAME = 'action'

    ACTION_CREATE = 'Create'
    ACTION_UPDATE = 'Update'
    ACTION_DELETE = 'Delete'
    ACTION_SET = 'Set'
    ACTION_UNSET = 'Unset'

    ACTION_CHOICES = (
        (ACTION_CREATE, _('Create')),
        (ACTION_UPDATE, _('Update')),
        (ACTION_DELETE, _('Delete')),
        (ACTION_SET, _('Set')),
        (ACTION_UNSET, _('Unset')),
    )

    ACTION_MAX_LENGTH = max(map(lambda choice: len(choice[0]), ACTION_CHOICES))

    DESCRIPTION_FIELD_NAME = 'description'

    OBJECT_FIELD_NAME = 'content'

    CREATED_AT_FIELD_NAME = 'created_at'

    UPDATED_AT_FIELD_NAME = 'updated_at'

    class Meta:  # type: ignore[override]
        abstract = True

    def get_revision(self) -> AbstractRevision:
        return getattr(self, self.__class__.REVISION_FIELD_NAME)

    def set_revision(  # type: ignore[override]
        self,
        revision: AbstractRevision,
    ) -> None:
        setattr(self, self.__class__.REVISION_FIELD_NAME, revision)

    def get_actor(self) -> Optional[models.Model]:
        return getattr(self, self.__class__.ACTOR_FIELD_NAME)

    def set_actor(
        self,
        actor: Optional[models.Model],
    ) -> None:
        setattr(self, self.__class__.ACTOR_FIELD_NAME, actor)

    def get_action(self) -> str:
        return getattr(self, self.__class__.ACTION_FIELD_NAME)

    def set_action(
        self,
        action: str,
    ) -> None:
        setattr(self, self.__class__.ACTION_FIELD_NAME, action)

    def get_description(self) -> str:
        return getattr(self, self.__class__.DESCRIPTION_FIELD_NAME)

    def set_description(
        self,
        description: str,
    ) -> None:
        setattr(self, self.__class__.DESCRIPTION_FIELD_NAME, description)

    def get_object(self) -> Model:
        return getattr(self, self.__class__.OBJECT_FIELD_NAME)

    def set_object(
        self,
        object_: Model,
    ) -> None:
        setattr(self, self.__class__.OBJECT_FIELD_NAME, object_)

    def get_created_at(self) -> datetime.datetime:
        return getattr(self, self.__class__.CREATED_AT_FIELD_NAME)

    def set_created_at(
        self,
        created_at: datetime.datetime,
    ) -> None:
        setattr(self, self.__class__.CREATED_AT_FIELD_NAME, created_at)

    def get_updated_at(self) -> datetime.datetime:
        return getattr(self, self.__class__.UPDATED_AT_FIELD_NAME)

    def set_updated_at(
        self,
        updated_at: datetime.datetime,
    ) -> None:
        setattr(self, self.__class__.UPDATED_AT_FIELD_NAME, updated_at)


class BaseDelta(
    AbstractDelta,
    models.Model,
):

    revision: _typing.ForeignKey[
        AbstractRevision,
        AbstractRevision,
    ] = models.ForeignKey(
        verbose_name=_('revision'),
        to=settings.REVISION_MODEL,
        on_delete=models.CASCADE,
        related_name='%(app_label)s_%(class)s_set',
        related_query_name='%(app_label)s_%(class)s',
        blank=False,
        null=False,
        db_column='revision_id',
    )

    actor_type: _typing.ForeignKey[
        Optional[ContentType],
        Optional[ContentType],
    ] = models.ForeignKey(
        verbose_name=_('actor type'),
        to=ContentType,
        on_delete=models.DO_NOTHING,
        related_name='%(app_label)s_%(class)s_set_created_by',
        related_query_name='%(app_label)s_%(class)s_created_by',
        blank=True,
        null=True,
        db_column='actor_type_id',
    )

    actor_id: _typing.TextField[
        Optional[str],
        Optional[str],
    ] = models.TextField(
        verbose_name=_('actor ID'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )

    actor = GenericForeignKey(
        ct_field='actor_type',
        fk_field='actor_id',
    )

    action: _typing.CharField[
        str,
        str,
    ] = models.CharField(
        verbose_name=_('action'),
        max_length=AbstractDelta.ACTION_MAX_LENGTH,
        choices=AbstractDelta.ACTION_CHOICES,
        blank=False,
        null=False,
        db_index=True,
    )

    description: _typing.TextField[
        str,
        str,
    ] = models.TextField(
        verbose_name=_('description'),
        blank=True,
        null=False,
        default='',
    )

    content_type: _typing.ForeignKey[
        ContentType,
        ContentType,
    ] = models.ForeignKey(
        verbose_name=_('content type'),
        to=ContentType,
        on_delete=models.DO_NOTHING,
        related_name='%(app_label)s_%(class)s_set',
        related_query_name='%(app_label)s_%(class)s',
        blank=True,
        null=True,
        db_column='content_type_id',
    )

    content_id: _typing.TextField[
        Optional[str],
        Optional[str],
    ] = models.TextField(
        verbose_name=_('content ID'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )

    content = GenericForeignKey(
        ct_field='content_type',
        fk_field='content_id',
    )

    created_at: _typing.DateTimeField[
        datetime.datetime,
        datetime.datetime,
    ] = models.DateTimeField(
        verbose_name=_('created at'),
        auto_now_add=True,
        blank=True,
        null=False,
        db_index=True,
    )

    updated_at: _typing.DateTimeField[
        datetime.datetime,
        datetime.datetime,
    ] = models.DateTimeField(
        verbose_name=_('updated at'),
        auto_now=True,
        blank=True,
        null=False,
        db_index=True,
    )

    class Meta:  # type: ignore[override]
        abstract = True


class Delta(
    BaseDelta,
    models.Model,
):

    revision: _typing.ForeignKey[
        AbstractRevision,
        AbstractRevision,
    ] = models.ForeignKey(
        verbose_name=_('revision'),
        to=settings.REVISION_MODEL,
        on_delete=models.CASCADE,
        related_name='deltas',
        related_query_name='delta',
        blank=False,
        null=False,
        db_column='revision_id',
    )

    actor_type: _typing.ForeignKey[
        Optional[ContentType],
        Optional[ContentType],
    ] = models.ForeignKey(
        verbose_name=_('actor type'),
        to=ContentType,
        on_delete=models.DO_NOTHING,
        related_name='deltas_created_by',
        related_query_name='delta_created_by',
        blank=True,
        null=True,
        db_column='actor_type_id',
    )

    content_type: _typing.ForeignKey[
        ContentType,
        ContentType,
    ] = models.ForeignKey(
        verbose_name=_('content type'),
        to=ContentType,
        on_delete=models.DO_NOTHING,
        related_name='deltas',
        related_query_name='delta',
        blank=True,
        null=True,
        db_column='content_type_id',
    )

    class Meta:  # type: ignore[override]

        verbose_name = _('delta')

        verbose_name_plural = _('deltas')

        db_table = 'revy__deltas'

        swappable = settings.DELTA_MODEL_ATTNAME


class AbstractObjectDelta(
    revy.abc.ObjectDelta,
    AbstractDelta,
    models.Model,
):

    PARENT_LINK_FIELD_NAME = 'object_delta_ptr'

    class Meta:  # type: ignore[override]
        abstract = True

    def get_attribute_deltas(self) -> _typing.QuerySet['AbstractAttributeDelta']:  # type: ignore[misc]
        attribute_delta_class = get_attribute_delta_model()
        attribute_deltas = attribute_delta_class.objects.filter(**{
            attribute_delta_class.PARENT_LINK_FIELD_NAME: self.pk,
        })
        return attribute_deltas

    def get_actors(self) -> Iterable[Model]:
        seen: Set[Model] = set()
        actor = self.get_actor()
        if actor is not None:
            yield actor
            seen.add(actor)
        for attribute_delta in self.get_attribute_deltas():
            actor = attribute_delta.get_actor()
            if actor is None:
                continue
            if actor in seen:
                continue
            yield actor
            seen.add(actor)


class BaseObjectDelta(
    AbstractObjectDelta,
    models.Model,
):

    class Meta:  # type: ignore[override]
        abstract = True


class ObjectDelta(
    BaseObjectDelta,
    Delta,
):

    object_delta_ptr: _typing.OneToOneField[
        AbstractDelta,
        AbstractDelta,
    ] = models.OneToOneField(
        verbose_name=_('object delta pointer'),
        to=settings.DELTA_MODEL,
        on_delete=models.CASCADE,
        related_name='object_delta',
        related_query_name='object_delta',
        parent_link=True,
    )

    class Meta:  # type: ignore[override]

        verbose_name = _('object delta')

        verbose_name_plural = _('object deltas')

        db_table = 'revy__object_deltas'

        swappable = settings.OBJECT_DELTA_MODEL_ATTNAME


class AbstractAttributeDelta(
    revy.abc.AttributeDelta,
    AbstractDelta,
    models.Model,
):

    PARENT_LINK_FIELD_NAME = 'attribute_delta_ptr'

    OBJECT_DELTA_FIELD_NAME = 'parent'

    ATTRIBUTE_NAME_FIELD_NAME = 'field_name'

    OLD_VALUE_FIELD_NAME = 'old_value'

    NEW_VALUE_FIELD_NAME = 'new_value'

    class Meta:  # type: ignore[override]
        abstract = True

    def get_object_delta(self) -> AbstractObjectDelta:
        return getattr(self, self.__class__.OBJECT_DELTA_FIELD_NAME)

    def set_object_delta(  # type: ignore[override]
        self,
        object_delta: AbstractObjectDelta,
    ) -> None:
        setattr(self, self.__class__.OBJECT_DELTA_FIELD_NAME, object_delta)

    def get_attribute_name(self) -> str:
        return getattr(self, self.__class__.ATTRIBUTE_NAME_FIELD_NAME)

    def set_attribute_name(
        self,
        attribute_name: str,
    ) -> None:
        setattr(self, self.__class__.ATTRIBUTE_NAME_FIELD_NAME, attribute_name)

    def get_old_value(self) -> Optional[Any]:
        return getattr(self, self.__class__.OLD_VALUE_FIELD_NAME)

    def set_old_value(
        self,
        old_value: Optional[Any],
    ) -> None:
        setattr(self, self.__class__.OLD_VALUE_FIELD_NAME, old_value)

    def get_new_value(self) -> Optional[Any]:
        return getattr(self, self.__class__.NEW_VALUE_FIELD_NAME)

    def set_new_value(
        self,
        new_value: Optional[Any],
    ) -> None:
        setattr(self, self.__class__.NEW_VALUE_FIELD_NAME, new_value)


class BaseAttributeDelta(
    AbstractAttributeDelta,
    models.Model,
):

    parent: _typing.ForeignKey[
        AbstractObjectDelta,
        AbstractObjectDelta,
    ] = models.ForeignKey(
        verbose_name=_('object delta'),
        to=settings.OBJECT_DELTA_MODEL,
        on_delete=models.CASCADE,
        related_name='child_%(app_label)s_%(class)s_set',
        related_query_name='child_%(app_label)s_%(class)s',
        blank=False,
        null=False,
        db_column='parent_id',
    )

    field_name: _typing.TextField[
        str,
        str,
    ] = models.TextField(
        verbose_name=_('field name'),
        blank=False,
        null=False,
        db_index=True,
    )

    old_value = models.JSONField(
        verbose_name=_('old value'),
        encoder=JSONEncoder,
        blank=True,
        null=True,
        default=None,
    )

    new_value = models.JSONField(
        verbose_name=_('new value'),
        encoder=JSONEncoder,
        blank=True,
        null=True,
        default=None,
    )

    class Meta:  # type: ignore[override]
        abstract = True


class AttributeDelta(
    BaseAttributeDelta,
    Delta,
):

    attribute_delta_ptr: _typing.OneToOneField[
        AbstractDelta,
        AbstractDelta,
    ] = models.OneToOneField(
        verbose_name=_('attribute delta'),
        to=settings.DELTA_MODEL,
        on_delete=models.CASCADE,
        related_name='attribute_delta',
        related_query_name='attribute_delta',
        parent_link=True,
    )

    parent: _typing.ForeignKey[
        AbstractObjectDelta,
        AbstractObjectDelta,
    ] = models.ForeignKey(
        verbose_name=_('parent'),
        to=settings.OBJECT_DELTA_MODEL,
        on_delete=models.CASCADE,
        related_name='children',
        related_query_name='child',
        blank=False,
        null=False,
        db_column='parent_id',
    )

    class Meta:  # type: ignore[override]

        verbose_name = _('attribute delta')

        verbose_name_plural = _('attribute deltas')

        db_table = 'revy__attribute_deltas'

        swappable = settings.ATTRIBUTE_DELTA_MODEL_ATTNAME


@dataclasses.dataclass()
class ModelInstanceState:

    instance: Model = dataclasses.field()

    is_initialized: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    init_keys: List[str] = dataclasses.field(
        kw_only=True,
        default_factory=list,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    initial_values: Dict[str, Any] = dataclasses.field(
        kw_only=True,
        default_factory=dict,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    previous_values: Dict[str, Any] = dataclasses.field(
        kw_only=True,
        default_factory=dict,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_new: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_being_saved: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_saved: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_being_fetched: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_fetched: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_being_deleted: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    is_deleted: bool = dataclasses.field(
        kw_only=True,
        default=False,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    attribute_deltas: List[AbstractAttributeDelta] = dataclasses.field(
        kw_only=True,
        default_factory=list,
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    field_name_to_attribute_delta_index_mapping: DefaultDict[str, List[int]] = dataclasses.field(
        kw_only=True,
        default_factory=lambda: defaultdict(list),
        metadata=wrap_metadata(
            snapshot=True,
        ),
    )

    STATE_ATTNAME = '__revy__state'

    @classmethod
    def get_for(
        cls,
        instance: Model,
    ) -> Optional['ModelInstanceState']:
        return getattr(
            instance,
            cls.STATE_ATTNAME,
            None,
        )

    @classmethod
    def get_or_create_for(
        cls,
        instance: Model,
    ) -> 'ModelInstanceState':
        state = cls.get_for(instance)
        if state is not None:
            return state
        return cls(instance)

    def __post_init__(self) -> None:
        if self.__class__.get_for(self.instance) is None:
            setattr(self.instance, self.__class__.STATE_ATTNAME, self)

    def backup(self) -> Dict[str, Any]:
        snapshot = dict()
        for field in dataclasses.fields(self):
            metadata = unwrap_metadata(field)
            if metadata.get('snapshot', False):
                snapshot[field.name] = getattr(self, field.name)
                if isinstance(snapshot[field.name], (list, dict)):
                    snapshot[field.name] = snapshot[field.name].copy()
        return snapshot

    def restore(
        self,
        snapshot: Dict[str, Any],
    ) -> None:
        for field in dataclasses.fields(self):
            metadata = unwrap_metadata(field)
            if metadata.get('snapshot', False):
                setattr(self, field.name, snapshot.get(field.name))


def get_model_instance_state(
    instance: Model,
) -> ModelInstanceState:
    return ModelInstanceState.get_or_create_for(instance)



from typing import TYPE_CHECKING

from django.db import models

from revy.contrib.django._typing_ext import mimic_generic


__all__ = (
    'IS_MYPY_DJANGO_PLUGIN_ENABLED',
    'CharField',
    'TextField',
    'DateTimeField',
    'ForeignKey',
    'OneToOneField',
    'QuerySet',
)


IS_MYPY_DJANGO_PLUGIN_ENABLED = False

try:
    import mypy_django_plugin  # noqa
    try:
        if TYPE_CHECKING:
            models.Field[object, object]  # noqa
            IS_MYPY_DJANGO_PLUGIN_ENABLED = True
    except TypeError:
        IS_MYPY_DJANGO_PLUGIN_ENABLED = False
except ImportError:
    IS_MYPY_DJANGO_PLUGIN_ENABLED = False


CharField = models.CharField
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    CharField = mimic_generic(models.CharField)  # type: ignore[misc]


TextField = models.TextField
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    TextField = mimic_generic(models.TextField)  # type: ignore[misc]


DateTimeField = models.DateTimeField
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    DateTimeField = mimic_generic(models.DateTimeField)  # type: ignore[misc]


ForeignKey = models.ForeignKey
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    ForeignKey = mimic_generic(models.ForeignKey)  # type: ignore[misc]


OneToOneField = models.OneToOneField
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    OneToOneField = mimic_generic(models.OneToOneField)  # type: ignore[misc]


QuerySet = models.QuerySet
if not TYPE_CHECKING or not IS_MYPY_DJANGO_PLUGIN_ENABLED:
    QuerySet = mimic_generic(models.QuerySet)  # type: ignore[misc]



from contextlib import suppress
from typing import (
    Any,
    List,
    Type,
    cast,
)

from django.apps import apps
from django.db.models import Model

from revy.contrib.django.conf import settings
from revy.contrib.django.patcher import Patcher


__all__ = (
    'setup',
)


_PATCHED_MODELS: List[Type[Model]] = []


def setup(
    *args: Any,
    **kwargs: Any,
) -> None:
    setting = kwargs.get('setting', '')
    if setting and not setting.startswith('REVY_'):
        return
    for patched_model in _PATCHED_MODELS:
        Patcher.unpatch_model(patched_model)
    _PATCHED_MODELS.clear()
    is_revy_installed = False
    with suppress(LookupError):
        revy_app_config = apps.get_app_config('revy')
        is_revy_installed = revy_app_config is not None
    if not is_revy_installed:
        return
    models: List[Type[Model]] = []
    if '*' in settings.MODELS:
        models = apps.get_models(
            include_auto_created=settings.INCLUDE_AUTO_CREATED_MODELS,
        )
    else:
        for model_name in settings.MODELS:
            model = cast(Type[Model], apps.get_model(model_name))
            models.append(model)
    for model in models:
        Patcher.patch_model(model)
        _PATCHED_MODELS.append(model)



from typing import List

from _core import DOCS_SOURCE_DIR_PATH


FILE_PATH = DOCS_SOURCE_DIR_PATH / 'api' / 'modules.rst'


def main() -> None:
    lines: List[str]
    with open(FILE_PATH, 'r') as f:
        lines = f.readlines()
    lines[0] = 'API Reference\n'
    lines[1] = f'{"=" * (len(lines[0]) - 1)}\n'
    with open(FILE_PATH, 'w') as f:
        f.writelines(lines)


if __name__ == '__main__':
    main()



import unittest

import revy


class ContextTestCase(unittest.TestCase):

    def test_simple_global_context_stack(self) -> None:
        with revy.Context():
            revy.Context.set_actor(1)

            with revy.Context():
                self.assertEqual(revy.Context.get_actor(), 1)
                revy.Context.set_actor(None)
                self.assertIsNone(revy.Context.get_actor())

            self.assertEqual(revy.Context.get_actor(), 1)
        self.assertIsNone(revy.Context.get_actor())

    def test_disable_enable(self) -> None:
        self.assertFalse(revy.Context.is_enabled())
        self.assertTrue(revy.Context.is_disabled())

        with revy.Context():
            self.assertFalse(revy.Context.is_disabled())
            self.assertTrue(revy.Context.is_enabled())

            with revy.Context():

                revy.Context.disable()

                self.assertFalse(revy.Context.is_enabled())
                self.assertTrue(revy.Context.is_disabled())

                with revy.Context():
                    self.assertFalse(revy.Context.is_enabled())
                    self.assertTrue(revy.Context.is_disabled())

                self.assertFalse(revy.Context.is_enabled())
                self.assertTrue(revy.Context.is_disabled())

            self.assertFalse(revy.Context.is_disabled())
            self.assertTrue(revy.Context.is_enabled())

        self.assertFalse(revy.Context.is_enabled())
        self.assertTrue(revy.Context.is_disabled())



import sys
from typing import (
    Optional,
    Sequence,
)

from revy._argparse import ArgumentParser


def main(
    argv: Optional[Sequence[str]] = None,
) -> int:
    argv = argv or sys.argv
    arg_parser = ArgumentParser()
    arg_parser.parse_known_args(argv)
    return 0


if __name__ == '__main__':
    status_code = main(argv=sys.argv)
    sys.exit(status_code)



import glob
import os
import subprocess

from _core import DIR_PATH


def main() -> None:
    for python_file_path in glob.glob(str(DIR_PATH / '*.py'), recursive=False):
        if python_file_path.split(os.sep)[-1].startswith('_'):
            continue
        subprocess.call(['python', python_file_path])


if __name__ == '__main__':
    main()



import unittest

import revy


class ASGIRefLocalStorageTestCase(unittest.TestCase):

    def test_get_base_context_class(self) -> None:
        self.assertIs(revy.ASGIRefLocalStorage.get_base_context_class(), revy.Context)

    def test_create_context_class(self) -> None:
        storage = revy.ASGIRefLocalStorage()
        context_class = storage.create_context_class()
        self.assertTrue(issubclass(context_class, revy.Context))



from typing import (
    Any,
    Iterable,
    TYPE_CHECKING,
)


if TYPE_CHECKING:
    from revy.abc.attribute_delta import AttributeDelta
    from revy.abc.delta import Delta
    from revy.abc.object_delta import ObjectDelta

__all__ = (
    'Revision',
)


class Revision:

    def get_description(self) -> str:
        raise NotImplementedError()

    def set_description(
        self,
        description: str,
    ) -> None:
        raise NotImplementedError()

    def get_deltas(self) -> Iterable['Delta']:
        raise NotImplementedError()

    def get_object_deltas(self) -> Iterable['ObjectDelta']:
        raise NotImplementedError()

    def get_attribute_deltas(self) -> Iterable['AttributeDelta']:
        raise NotImplementedError()

    def get_actors(self) -> Iterable[Any]:
        raise NotImplementedError()



from typing import (
    Any,
    TYPE_CHECKING,
    Type,
    TypeVar,
    cast,
)

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import (
    Field,
    JSONField,
    Model,
    OuterRef,
    Subquery,
)
from django.db.models.functions import JSONObject  # type: ignore[attr-defined]
from django.db.models.functions import Cast
from django.db.models.options import Options

from revy.contrib.django.utils import (
    get_attribute_delta_model,
    get_object_delta_model,
)


if TYPE_CHECKING:
    from revy.contrib.django.models import AbstractDelta


__all__ = (
    'ObjectSnapshot',
)


T = TypeVar('T')

IF_T = TypeVar('IF_T', bound='_InstanceField')


class _InstanceField(JSONField):

    model: Type[Model]

    def __init__(
        self,
        model: Type[Model],
        **kwargs: Any,
    ) -> None:
        self.model = model
        super(_InstanceField, self).__init__(**kwargs)

    def from_db_value(
        self: IF_T,
        value: Any,
        expression: IF_T,
        connection: BaseDatabaseWrapper,
    ) -> Any:
        python_value = super(_InstanceField, self).from_db_value(  # type: ignore[misc]
            value,
            expression,
            connection,
        )
        if isinstance(python_value, dict):
            return self.model(**python_value)
        if isinstance(python_value, (list, tuple)):
            container_type = type(python_value)
            return container_type(map(lambda kwargs: self.model(**kwargs), python_value))
        return python_value


class ObjectSnapshot(Subquery):

    @classmethod
    def get_ct_fk_fields(
        cls,
        delta_model: Type["AbstractDelta"],
    ) -> tuple[Field, Field]:
        options = cast(Options, delta_model._meta)  # noqa
        generic_fk_field = cast(
            GenericForeignKey,
            options.get_field(delta_model.OBJECT_FIELD_NAME),
        )
        ct_field = cast(Field, options.get_field(generic_fk_field.ct_field))
        fk_field = cast(Field, options.get_field(generic_fk_field.fk_field))
        return ct_field, fk_field

    @classmethod
    def build_json_object(
        cls,
        model: Type[Model],
    ) -> JSONObject:
        object_delta_model = get_object_delta_model()
        attribute_delta_model = get_attribute_delta_model()
        annotations = dict()
        options = cast(Options, model._meta)  # noqa
        obj_ct_field, obj_fk_field = cls.get_ct_fk_fields(object_delta_model)
        att_ct_field, att_fk_field = cls.get_ct_fk_fields(attribute_delta_model)
        for field in options.get_fields():
            attname = getattr(field, 'attname', None)
            one_to_many = getattr(field, 'one_to_many', False)
            many_to_many = getattr(field, 'many_to_many', False)
            any_to_many = one_to_many or many_to_many
            if not attname or any_to_many:
                continue
            annotations[attname] = attribute_delta_model.objects.filter(
                **{
                    att_ct_field.attname: OuterRef(obj_ct_field.attname),
                    att_fk_field.attname: OuterRef(obj_fk_field.attname),
                    attribute_delta_model.ATTRIBUTE_NAME_FIELD_NAME: attname,
                    f'{attribute_delta_model.OBJECT_DELTA_FIELD_NAME}__pk__lte': OuterRef('pk'),
                },
            ).order_by(
                '-pk',
            ).values(
                attribute_delta_model.NEW_VALUE_FIELD_NAME,
            )[:1]
        return JSONObject(**annotations)

    @classmethod
    def build_instance(
        cls,
        model: Type[Model],
    ) -> Cast:
        return Cast(cls.build_json_object(model), output_field=_InstanceField(model))

    def __init__(
        self,
        model: Type[Model],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        object_delta_model = get_object_delta_model()
        ContentType.objects.get_for_model(model)
        queryset = object_delta_model.objects.filter(
            pk=OuterRef('pk'),
        ).annotate(
            snapshot=self.__class__.build_instance(model),
        ).values(
            'snapshot',
        )[:1]
        super(ObjectSnapshot, self).__init__(queryset, **kwargs)



from typing import (
    Any,
    Generic,
    Type,
    TypeVar,
    cast,
)


__all__ = (
    'mimic_generic',
)


T = TypeVar('T')


class _Generic(
    Generic[T],
):

    type_: Type[T]

    def __init__(
        self,
        type_: Type[T],
    ) -> None:
        self.type_ = type_

    def __getitem__(
        self,
        *args: Any,
    ) -> Type[T]:
        return self.type_


def mimic_generic(
    type_: Type[T],
) -> Type[T]:
    return cast(Type[T], _Generic(type_))



from typing import (
    Any,
    Callable,
    Iterator,
    Optional,
)

from django.db.models import (
    Model,
    QuerySet,
)
from django.db.models.deletion import (
    CASCADE as _CASCADE,
    Collector,
    SET as _SET,
    SET_DEFAULT as _SET_DEFAULT,
    SET_NULL as _SET_NULL,
)
from django.db.models.fields.related import RelatedField

from revy.contrib.django.conf import settings
from revy.contrib.django.utils import get_context_class


__all__ = (
    'CASCADE',
    'SET',
    'SET_NULL',
    'SET_DEFAULT',
)


Context = get_context_class()


def _get_sub_object_iterator(
    sub_objects: QuerySet,
) -> Iterator[Model]:
    return sub_objects.iterator(
        chunk_size=settings.DELETION_ITERATOR_CHUNK_SIZE,
    )


def CASCADE(  # noqa
    collector: Collector,
    field: RelatedField,
    sub_objects: QuerySet,
    using: str,
) -> None:

    if Context.is_disabled():
        return _CASCADE(collector, field, sub_objects, using)

    with Context.via_actor(None):
        sub_object_iterator = _get_sub_object_iterator(sub_objects)
        for sub_object in sub_object_iterator:
            with Context.via_object_delta_description(
                Context.get_deletion_description(),
            ):
                sub_object.delete(using=using)


def SET(  # noqa
    value: Optional[Any],
) -> Callable[
    [Collector, RelatedField, QuerySet, str],
    None,
]:

    def set_on_delete(
        collector: Collector,
        field: RelatedField,
        sub_objects: QuerySet,
        using: str,
    ) -> None:

        if Context.is_disabled():
            return _SET(value)(collector, field, sub_objects, using)

        new_value = value
        if callable(new_value):
            new_value = new_value()

        with Context.via_actor(None):
            sub_object_iterator = _get_sub_object_iterator(sub_objects)
            for sub_object in sub_object_iterator:
                with Context.via_attribute_delta_description(
                    Context.get_deletion_description(),
                ):
                    setattr(sub_object, field.attname, new_value)
                sub_object.save(using=using)

    setattr(
        set_on_delete,
        'deconstruct',
        lambda: ('revy.contrib.django.models.SET', (value,), {}),
    )

    return set_on_delete


def SET_NULL(  # noqa
    collector: Collector,
    field: RelatedField,
    sub_objects: QuerySet,
    using: str,
) -> None:

    if Context.is_disabled():
        return _SET_NULL(collector, field, sub_objects, using)

    with Context.via_actor(None):
        sub_object_iterator = _get_sub_object_iterator(sub_objects)
        for sub_object in sub_object_iterator:
            with Context.via_attribute_delta_description(
                Context.get_deletion_description(),
            ):
                setattr(sub_object, field.attname, None)
            sub_object.save(using=using)


def SET_DEFAULT(  # noqa
    collector: Collector,
    field: RelatedField,
    sub_objects: QuerySet,
    using: str,
) -> None:

    if Context.is_disabled():
        return _SET_DEFAULT(collector, field, sub_objects, using)

    with Context.via_actor(None):
        sub_object_iterator = _get_sub_object_iterator(sub_objects)
        for sub_object in sub_object_iterator:
            with Context.via_attribute_delta_description(
                Context.get_deletion_description(),
            ):
                setattr(sub_object, field.attname, field.get_default())
            sub_object.save(using=using)



import argparse
from typing import Any

from revy.__version__ import __version__


__all__ = (
    'ArgumentParser',
)


class ArgumentParser(argparse.ArgumentParser):

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        kwargs.setdefault('prog', 'revy')
        kwargs.setdefault('description', 'A toolkit for building data-driven version control systems.')
        super(ArgumentParser, self).__init__(*args, **kwargs)
        self.add_argument(
            '-v',
            '--version',
            action='version',
            version=f'%(prog)s {__version__}',
        )



import abc
from typing import (
    TYPE_CHECKING,
    Type,
)

import stackholm


if TYPE_CHECKING:
    from revy.context import Context


__all__ = (
    'Storage',
    'ASGIRefLocalStorage',
)


class Storage(
    stackholm.Storage,
    metaclass=abc.ABCMeta,
):

    @classmethod
    def get_base_context_class(cls) -> Type['Context']:
        from revy.context import Context
        return Context


class ASGIRefLocalStorage(
    Storage,
    stackholm.ASGIRefLocalStorage,
):
    ...



from typing import Any

from revy.abc.revision import Revision


__all__ = (
    'Delta',
)


class Delta:

    def get_revision(self) -> Revision:
        raise NotImplementedError()

    def set_revision(
        self,
        revision: Revision,
    ) -> None:
        raise NotImplementedError()

    def get_actor(self) -> Any:
        raise NotImplementedError()

    def set_actor(
        self,
        actor: Any
    ) -> None:
        raise NotImplementedError()

    def get_action(self) -> str:
        raise NotImplementedError()

    def set_action(
        self,
        action: str
    ) -> None:
        raise NotImplementedError()

    def get_object(self) -> Any:
        raise NotImplementedError()

    def set_object(
        self,
        object_: Any,
    ) -> None:
        raise NotImplementedError()

    def get_description(self) -> str:
        raise NotImplementedError()

    def set_description(
        self,
        description: str
    ) -> None:
        raise NotImplementedError()
